

/***********************************************************************
 *                                                                     *
 * This file was generated automatically. Do not edit.                 *
 *                                                                     *
 ***********************************************************************/


#ifndef __README_H
#define __README_H

#define README_SECTION_installationandconfiguration "Part I"
#define README_TITLE_installationandconfiguration "Installation and Configuration Instructions"

#define README_SECTION_introduction "Chapter 1"
#define README_TITLE_introduction "Introduction"

#define README_SECTION_minimumrequirements "Chapter 2"
#define README_TITLE_minimumrequirements "Minimum Software Requirements"

#define README_SECTION_selectdriver "Chapter 3"
#define README_TITLE_selectdriver "Selecting and Downloading the NVIDIA Packages for Your System"

#define README_SECTION_installdriver "Chapter 4"
#define README_TITLE_installdriver "Installing the NVIDIA Driver"

#define README_SECTION_installedcomponents "Chapter 5"
#define README_TITLE_installedcomponents "Listing of Installed Components"

#define README_SECTION_editxconfig "Chapter 6"
#define README_TITLE_editxconfig "Configuring X for the NVIDIA Driver"

#define README_SECTION_faq "Chapter 7"
#define README_TITLE_faq "Frequently Asked Questions"

#define README_SECTION_commonproblems "Chapter 8"
#define README_TITLE_commonproblems "Common Problems"

#define README_SECTION_knownissues "Chapter 9"
#define README_TITLE_knownissues "Known Issues"

#define README_SECTION_dma_issues "Chapter 10"
#define README_TITLE_dma_issues "Allocating DMA Buffers on 64-bit Platforms"

#define README_SECTION_openglenvvariables "Chapter 11"
#define README_TITLE_openglenvvariables "Specifying OpenGL Environment Variable Settings"

#define README_SECTION_configtwinview "Chapter 12"
#define README_TITLE_configtwinview "Configuring Multiple Display Devices on One X Screen"

#define README_SECTION_xineramaglx "Chapter 13"
#define README_TITLE_xineramaglx "Configuring GLX in Xinerama"

#define README_SECTION_configmultxscreens "Chapter 14"
#define README_TITLE_configmultxscreens "Configuring Multiple X Screens on One Card"

#define README_SECTION_configtvout "Chapter 15"
#define README_TITLE_configtvout "Configuring TV-Out"

#define README_SECTION_xrandrextension "Chapter 16"
#define README_TITLE_xrandrextension "Support for the X Resize and Rotate Extension"

#define README_SECTION_configlaptop "Chapter 17"
#define README_TITLE_configlaptop "Configuring a Notebook"

#define README_SECTION_optimus "Chapter 18"
#define README_TITLE_optimus "Using the NVIDIA Driver with Optimus Laptops"

#define README_SECTION_programmingmodes "Chapter 19"
#define README_TITLE_programmingmodes "Programming Modes"

#define README_SECTION_flippingubb "Chapter 20"
#define README_TITLE_flippingubb "Configuring Flipping and UBB"

#define README_SECTION_procinterface "Chapter 21"
#define README_TITLE_procinterface "Using the Proc Filesystem Interface"

#define README_SECTION_powermanagement "Chapter 22"
#define README_TITLE_powermanagement "Configuring Power Management Support"

#define README_SECTION_xcompositeextension "Chapter 23"
#define README_TITLE_xcompositeextension "Using the X Composite Extension"

#define README_SECTION_nvidiasettings "Chapter 24"
#define README_TITLE_nvidiasettings "Using the nvidia-settings Utility"

#define README_SECTION_nvidia_smi "Chapter 25"
#define README_TITLE_nvidia_smi "Using the nvidia-smi Utility"

#define README_SECTION_nvidia_ml "Chapter 26"
#define README_TITLE_nvidia_ml "The NVIDIA Management Library"

#define README_SECTION_nvidia_debugdump "Chapter 27"
#define README_TITLE_nvidia_debugdump "Using the nvidia-debugdump Utility"

#define README_SECTION_nvidia_persistenced "Chapter 28"
#define README_TITLE_nvidia_persistenced "Using the nvidia-persistenced Utility"

#define README_SECTION_sli "Chapter 29"
#define README_TITLE_sli "Configuring SLI and Multi-GPU FrameRendering"

#define README_SECTION_framelock "Chapter 30"
#define README_TITLE_framelock "Configuring Frame Lock and Genlock"

#define README_SECTION_sdi "Chapter 31"
#define README_TITLE_sdi "Configuring SDI Video Output"

#define README_SECTION_depth30 "Chapter 32"
#define README_TITLE_depth30 "Configuring Depth 30 Displays"

#define README_SECTION_randr14 "Chapter 33"
#define README_TITLE_randr14 "Offloading Graphics Display with RandR 1.4"

#define README_SECTION_addressingcapabilities "Chapter 34"
#define README_TITLE_addressingcapabilities "Addressing Capabilities"

#define README_SECTION_addtlresources "Chapter 35"
#define README_TITLE_addtlresources "NVIDIA Contact Info and Additional Resources"

#define README_SECTION_acknowledgements "Chapter 36"
#define README_TITLE_acknowledgements "Acknowledgements"


#define README_SECTION_appendices "Part II"
#define README_TITLE_appendices "Appendices"

#define README_SECTION_supportedchips "Appendix A"
#define README_TITLE_supportedchips "Supported NVIDIA GPU Products"

#define README_SECTION_xconfigoptions "Appendix B"
#define README_TITLE_xconfigoptions "X Config Options"

#define README_SECTION_displaydevicenames "Appendix C"
#define README_TITLE_displaydevicenames "Display Device Names"

#define README_SECTION_glxsupport "Appendix D"
#define README_TITLE_glxsupport "GLX Support"

#define README_SECTION_dpi "Appendix E"
#define README_TITLE_dpi "Dots Per Inch"

#define README_SECTION_i2c "Appendix F"
#define README_TITLE_i2c "i2c Bus Support"

#define README_SECTION_vdpausupport "Appendix G"
#define README_TITLE_vdpausupport "VDPAU Support"

#define README_SECTION_audiosupport "Appendix H"
#define README_TITLE_audiosupport "Audio Support"

#define README_SECTION_newusertips "Appendix I"
#define README_TITLE_newusertips "Tips for New Linux Users"

#define README_SECTION_profiles "Appendix J"
#define README_TITLE_profiles "Application Profiles"

#define README_SECTION_gpunames "Appendix K"
#define README_TITLE_gpunames "GPU Names"



#endif

